#include "rtc.h"

#define TRUE	1
#define FALSE	0
#define	AT_CNTL		0x70		/* AT CMOS, clock control register.	*/
#define AT_DATA		0x71		/* AT CMOS, clock data register.	*/
#define CLK_HOLD	0x0080		/* Value used to set bit 7 of Reg B */
#define CLK_START	0x007f		/* Used to Reset( 0 ) bit 7 of Reg B */
#define H2412		0x0001		/* Used to force 24 hour format		*/
#define REG_A		0x0A		/* MC146818 Status Register A		*/
#define	REG_B		0x0B		/* MC146818 Status Register B		*/
#define CMOS_CNT	0xffff		/* Number of retries when testing Reg. A */

extern	unsigned	date[2], cent, year, month, day, hours, minutes, seconds;
extern	unsigned	int_bcd(), bcd_int();

static	unsigned cmos(i)
unsigned	i;
{
	unsigned temp;

	_disable();			/* cli											*/
	outp(0x70,i);		/* mov ax, 14[bp] , b out 70h (CMOS ADDR REG)	*/
						/* jmp . 										*/
	temp = inp(0x71);	/* b in 71h (CMOS Data Register)				*/
	_enable();			/* sti											*/
	return(temp);
}


static unsigned cmos_set(i, d)
unsigned	i,d;
{
	_disable();			/* cli 											*/
	outp(0x70,i);		/* mov ax, 14[bp], b out 70h (CMOS offset)		*/
						/* jmp .										*/
	outp(0x71,d);		/* mov ax,16[bp], b out 71h (set time/date val) */
	_enable();			/* sti											*/
	return(0);
}

at(int setflag)
{
register	unsigned	x = 0;
register	unsigned  count;
unsigned	clk_buf[2][7];
int			i, z, loop_flag;

#ifdef __QNXNTO__
#if _NTO_VERSION >= 110
	ThreadCtl(_NTO_TCTL_IO,NULL);
#else
	ThreadPriv(_NTO_TPRIV_IO,NULL);   
#endif
#endif


	if(setflag) { 			/* Set hardware clock.	*/
		seconds	= int_bcd(seconds);
		minutes	= int_bcd(minutes);
		hours	= int_bcd(hours);
		day 	= int_bcd(day);
		month	= int_bcd(month);
		if ( year > 99 ) year -= 100;
#ifdef DIAG
		fprintf(stderr,"rtc set: cent=%d; year=%d (after adjustment)\n",cent,year);
#endif

		year	= int_bcd(year);
		cent	= int_bcd(cent);

		x = cmos( REG_A );
		x = x & 0xf | 0x20;	/* Set internal ocillator running */
		cmos_set( REG_A, x );
		
		x = cmos( REG_B );
		x |= CLK_HOLD | H2412;	/* Set bit 7 & 1 of Status Register B to a '1',   */
		cmos_set( REG_B, x );	/* Hold clock updates, clear UIP bit in REG A */ 
		cmos_set(7, day);
		cmos_set(8, month);
		cmos_set(9, year);

#ifdef NEVER
  if (use_century) cmos_set(50,cent);
#endif
 		cmos_set(4, hours);
		cmos_set(2, minutes);
		cmos_set(0, seconds);
		x &= CLK_START;
		cmos_set( REG_B, x );	/* Restart the clock.	*/
	}else {					/* Get the time from the hardware clock.	*/
		x = cmos( REG_A ) & 0x60;
		if(x != 0x20) { 
			fprintf( stderr, "RTC: Clock not running, Must be set first 0x%x\n", x);
			exit(-1);
		}

		do {
			for ( x = 0, z = 1 ; (z <= 2) && !(x & 0x0080) ; z++ ) 	
				for ( count=CMOS_CNT ; !(x & 0x0080) && count ; count-- ) {
					outp( AT_CNTL, REG_A );
					x =inp( AT_DATA );	/* Test bit 7 of Reg. A, ( == 1 ?)	*/
				}
			
			for ( x=0x080, count = CMOS_CNT ; (x & 0x0080) && count ; count-- ) {
				outp( AT_CNTL, REG_A );
				x =inp( AT_DATA );  /* Test bit 7 of Reg. A, ( == 0 ?)	 */
			}

			if ( count == 0 ) {
				fprintf( stderr, "RTC: cannot read AT clock\n");
				exit(-1);
			}
				
			clk_buf[0][0] = cmos(0);		/* seconds	*/
			clk_buf[0][1] = cmos(2);		/* minutes	*/
			clk_buf[0][2] = cmos(4);		/* hours	*/
			clk_buf[0][3] = cmos(7);		/* day		*/
			clk_buf[0][4] = cmos(8);		/* month	*/
			clk_buf[0][5] = cmos(9);		/* year		*/

#ifdef NEVER
			if (use_century)
				clk_buf[0][6] = cmos(50);		/* century  */
#endif

			clk_buf[1][0] = cmos(0);		/* seconds	*/
			clk_buf[1][1] = cmos(2);		/* minutes	*/
			clk_buf[1][2] = cmos(4);		/* hours	*/
			clk_buf[1][3] = cmos(7);		/* day		*/
			clk_buf[1][4] = cmos(8);		/* month	*/
			clk_buf[1][5] = cmos(9);		/* year		*/

			x = cmos( REG_B );
#ifdef NEVER
			if (use_century)
				clk_buf[1][6] = cmos(50);		/* century  */
#endif

			for ( loop_flag = TRUE, i=0 ; i <= 5 ; i++ )
				if ( clk_buf[0][i] != clk_buf[1][i] ) {
					loop_flag = FALSE;
					break;
					}

		 } while ( loop_flag == FALSE );

		seconds = bcd_int(clk_buf[0][0]);
		minutes = bcd_int(clk_buf[0][1]);

		hours   = bcd_int(clk_buf[0][2]);
		if (!(x & H2412) && (clk_buf[0][2] & 0x80)) {	 /* 12 hour format, PM */
			hours = bcd_int(clk_buf[0][2] & 0xff);
			hours += 12;
		}
		day		= bcd_int(clk_buf[0][3]);
		month	= bcd_int(clk_buf[0][4]);
		year	= bcd_int(clk_buf[0][5]);

#ifdef VERBOSE_SUPPORTED
		/* print diagnostics on read if double -v (-vv) specified */
		if (verbose>1) {
			fprintf(stdout,"cmos(0) seconds=%d\n",seconds);
			fprintf(stdout,"cmos(2) minutes=%d\n",minutes);
			fprintf(stdout,"cmos(4) hour   =%d\n",hours);
			fprintf(stdout,"cmos(7) day    =%d\n",day);
			fprintf(stdout,"cmos(8) month  =%d\n",month);
			fprintf(stdout,"cmos(9) year   =%d\n",year);
			fprintf(stdout,"(unused: cmos(50) century = %d)\n",bcd_int(cmos(50)));
		}
#endif

#ifdef NEVER
		if (use_century) {
			cent	= bcd_int(clk_buf[0][6]);
#ifdef DIAG
			fprintf(stderr,"rtc read: cent=%d; year=%d (b4 adjustment)\n",cent,year);
#endif
			if ( cent == 20 ) year += 100;
		}
#endif
		if (year<39) year+=100;
        }

	return(0);
	}
