#include "rtc.h"

#ifdef VERBOSE_SUPPORTED
#define VERBOSE_OPTIONS "v"
#else
#define VERBOSE_OPTIONS
#endif

#ifdef SLOWADJUST
#define SLOWADJUST_OPTIONS "r:S:"
#else
#define SLOWADJUST_OPTIONS
#endif

#ifdef NETCLOCK
#define NETCLOCK_STRING "|net"
#else
#define NETCLOCK_STRING
#endif

#define TXT(s) s
#define T_CANT_GET_OS_INFO	"rtc: Can't determine machine type!\n"
#define T_UNKNOWN_MACHINE_TYPE	"rtc: Unknown machine type (%s)\n"

#ifdef NEVER
#define USE_CENTURY (1==1)
#endif

extern int at (int setflag);
extern int ds1386(int setflag, unsigned base);
extern int rtc72423(int setflag, unsigned base);
#ifdef NETCLOCK
extern int net (int setflag);
#endif

unsigned	date[2], cent, year, month, day, hours, minutes, seconds;
unsigned	msec, dow;
#ifdef NETCLOCK
nid_t		node;
#endif
time_t		timevar;	/* used instead of the above for clock type net */
#ifdef SLOWADJUST
long		rate = 100;
long        maxsec = 60;
#endif
#ifdef VERBOSE_SUPPORTED
int			verbose=0;
#endif

unsigned	int_bcd(A)
unsigned	A;
{
	return( ((A/10) << 4) + (A % 10) );
}

unsigned	bcd_int(A)
unsigned	A;
{
	return( (A >> 4)*10 + (A & 0x0f) );
}

static	char *clock_names[] ={
			"net",   /* 0; do not remove; the index of these are hard-coded */
			"hw",    /* 1 */
			"at",    /* 2 */
			"ps/2",  /* 3 */
			"ds1386", /* 4 */
			"rtc72423", /* 5 */
			0
			};      	

#ifdef __USAGE
%C	[-l] [-r rate] [-s] [-S seconds] <clock_type>
Options:
 -l         Hardware clock is in local time (instead of UTC).
 -r rate    Rate at which to adjust the time if within 60 sec. (default 100)
 -s         Set clock to current OS time.
 -S seconds Maximum number of seconds difference between current
            time and new time before jam loading rather than
            slowly adjusting (see -r to adjust convergence speed). (default 60)
 -b         base address for DS1386 RAMified TIMEKEEPER, or FOX RTC-72423.
Where:
    <clock_type> is
        hw                 (hardware clock, local node)
        net <node>         (OS time on <node>)

    <clock_type> may also be
        at                 (ignore the Proc machine type, assume AT hardware)
        ps/2               (ignore the Proc machine type, assume PS/2 hardware)
        ds1386             (embedded system, use Dallas ds1386 chip)
        rtc72423           (embedded system, use Foxs RTC-72423)
    which is useful when running rtc under a Proc which knows about a
    previously unrecognized machine type or when the -M option has been
    specified to Proc.
Note:
    -l affects the time in both reading and writing, except
       when the time is being read from clock type 'net', in which
       case it has no effect.
    -s in conjunction with a net clock type will cause the
       date on the specified node to be changed to the current OS
       time of the node on which this command is executed.
    -r rate at which to slow down or speed up the clock until
       it catches up. The rate is turned into a percentage of the
       form 1/rate. The default rate of 100 is 1 percent (1/100).
#endif

#define TXT(s)	s

main(argc, argv)
int argc;
char *argv[];
	{
	int i,setflag=0, local=0;
	struct	tm *p,local_time;
	int opt, error=0;
	unsigned base;
#ifdef NETCLOCK
	int  net_flag=0;
#endif


	while ((opt=getopt(argc,argv,"b:ls" SLOWADJUST_OPTIONS VERBOSE_OPTIONS))!=-1)
	{
		switch(opt)
		{
			case 'b': base = strtol(optarg, 0, 16);	break;
			case 'l':
				local++;
				break;
#ifdef SLOWADJUST
			case 'r': rate = atol(optarg);	break;
			case 'S': maxsec = atol(optarg); break;
#endif
			case 's':
				setflag++;
				break;
#ifdef VERBOSE_SUPPORTED
            case 'v': verbose++; break;
#endif
			default:
				error++;
				break;
		}
	}

	if ((argc-optind)>2) {
		error++;
		fprintf(stderr,"rtc: Too many command line operands.\n");
	} else if ((argc-optind)<1) {
		error++;
		fprintf(stderr,"rtc: Must specify a clock type (hw"NETCLOCK_STRING"|at|ps/2|ds1386).\n");
	}
	
	if (geteuid()) {
		error++;
		fprintf(stderr,"rtc: operation not permitted -- must be root.\n");
	}

	if (error) exit(EXIT_FAILURE);

/*
 *	If -s option is true, read & decompose the QNX System Date.
 */
	if (setflag) {
		time(&timevar);		/* Get UTC	time.				*/
		/*
			printf("time = %ld\n",timevar);
		*/
		if (local)
			p = localtime(&timevar);	/* Break-down into localtime.	*/
		else	
		    p = gmtime(&timevar);		/* Break-down into UTC time.	*/

		day 	= p->tm_mday;
		month 	= p->tm_mon + (1);
		year 	= p->tm_year;
		cent	= ( year > 99 ) ? 20 : 19;
		hours 	= p->tm_hour;		
		minutes = p->tm_min;
		seconds = p->tm_sec;
		dow		= p->tm_wday;

		#ifdef DIAG
			printf("(setflag true) %d %d/%d/%d %d:%d:%d\n",cent,year,month,day,hours,minutes,seconds);
		#endif
		#ifdef VERBOSE_SUPPORTED
		if (verbose)
			printf("rtc set: current OS time is %d %d/%d/%d %d:%d:%d %s\n",cent,year,month,day,hours,minutes,seconds,
	                       local?"(Local)":"(UTC)");
		#endif
	}


	for(i = 0 ; clock_names[i] ; ++i) {
		if(strcmp(clock_names[i], argv[optind]) == 0) break;
	}
	
	if (!clock_names[i]) i=-1;

	if (i==1) {				/* 'hw' */
#ifndef __QNXNTO__
		struct _osinfo osdata;
		if (qnx_osinfo(0,&osdata) == -1) {
			perror(TXT(T_CANT_GET_OS_INFO));
			exit(EXIT_FAILURE);
		}
		if (!strcmp(osdata.machine,"AT") || !strcmp(osdata.machine,"PCI") ||!strcmp(osdata.machine,"EISA") ) i = 2;
		else if (!strcmp(osdata.machine,"PS/2")) i = 3;
		else {
			fprintf(stderr,TXT(T_UNKNOWN_MACHINE_TYPE),osdata.machine);
			exit(EXIT_FAILURE);
		}
#else
		i=2;   /* regular AT clock */		
#endif
	}

	
	switch(i)
	{
#ifdef NETCLOCK
		case 0 : 
			/* argv[optind] = 'net', argv[optind+1] = node number */
			if ((optind+1)<argc) {
				node = atoi(argv[optind+1]);
			} else node = 1;
			/*	printf("net %d\n",node); */
			net( setflag );
			net_flag++;
			break;
#endif

		case 2 :
		case 3 : /* both cases are the same now that the century byte is ignored */
			at(setflag);
			break;

		case 4 :
			ds1386(setflag, base);
			break;

		case 5 :
			rtc72423(setflag, base);
			break;

		default:
			fprintf(stderr, "rtc: %s - Unknown clock type. Must be hw"NETCLOCK_STRING"|at|ps/2.\n", argv[optind]);
			exit(EXIT_FAILURE);
	}

	
/* 
 * 	If -s option is not true, load the QNX System Date.
 */
    if ( !setflag ) {
		
#ifdef NETCLOCK
		if (!net_flag)
#endif
		{
			p = &local_time;
			p->tm_mday	= day;				/* Setup TM structure.	*/
			p->tm_mon	= month - (1);
			p->tm_year	= year;
			p->tm_hour	= hours;		
			p->tm_min	= minutes;
			p->tm_sec	= seconds;
			p->tm_isdst	= local?-1:0;

	
			#ifdef DIAG
			printf("(no setflag)%d/%d/%d %d:%d:%d\n",year,month,day,hours,minutes,seconds);
			printf("tm_isdst = %d\n",p->tm_isdst);
			printf("timezone = %d\n",timezone);
            #endif
			#ifdef VERBOSE_SUPPORTED
			if (verbose)
				printf("rtc read: hw clock reads (century=%d) %d/%d/%d %d:%d:%d %s\n",cent,year,month,day,hours,minutes,seconds,
	                       local?"(Local)":"(UTC)");
			#endif
			
			/*
			 *	MKTIME() assumes local time and will
			 *	subtract timezone to obtain UTC
			 */

			timevar = mktime(p);

			#ifdef DIAG
			printf("tm_isdst = %d\n",p->tm_isdst);
			printf("timezone = %d\n",timezone);
			printf("timevar = mktime(p) = %ld\n",timevar);
			#endif
		
			/* 
			 *	If hardware time was already in UTC, then we
			 *	overcompensated and must re-adjust to UTC.
			 */
	
			if ( !local )  timevar -= timezone;
		}

		/* Set QNX UTC time.	*/
		{
			struct timespec new, cur;
			long sec, usec;
			int rc = -1;
			long int c,d;

			new.tv_nsec = msec;
			new.tv_sec = timevar;

#ifdef SLOWADJUST
#ifndef __QNXNTO__
			getclock(TIMEOFDAY, &cur);

			sec = new.tv_sec - cur.tv_sec;
			if(sec > -maxsec &&  sec < maxsec) {
				usec  = new.tv_sec*1000000 + new.tv_nsec/1000;
				usec -= cur.tv_sec*1000000 + cur.tv_nsec/1000;
#ifdef VERBOSE_SUPPORTED
				if (verbose) {
					if (!qnx_adj_time(0,0,&c,&d)) {
						if (d) {
							fprintf(stderr,"Time is currently being adjusted %d ns/tick, %d ticks remaining\n",c,d);
							fprintf(stderr,"Will override.\n");
						}
					}
				}
#endif

				rc = qnx_adj_time(usec, rate, NULL, NULL);
	
#ifdef VERBOSE_SUPPORTED
				if (verbose)
				do {
					sleep(1);
					if (!qnx_adj_time(0,0,&c,&d)) {
						fprintf(stderr,"%d ns/tick, %d ticks remaining\n",c,d);
					} else break;
				} while (d!=0);			
#endif

				}
#else
            /* Neutrino version */

			clock_gettime(CLOCK_REALTIME,&cur);

			/* under nto this is more complicated. You can't just specify 
               the total time to be adjusted, rather, you must specify
               an amount per tick and a number of ticks (the product of
               which is the total time). This means that some sort of
               factoring algorithm must be used to produce an nsec
               per-tick amount and a number of ticks which will equal
               the total time required. */
               
			if((sec = new.tv_sec - cur.tv_sec) >= -maxsec  &&  sec <= maxsec) {
				struct _clockadjust adj;

#ifdef VERBOSE_SUPPORTED
				if (verbose) printf("rtc read: attempting slow adjust (delta = %d sec.)\n",sec);
#endif
				usec  = new.tv_sec*1000000 + new.tv_nsec/1000;
				usec -= cur.tv_sec*1000000 + cur.tv_nsec/1000;
    			
				/* we want usec*1000 total ns over 'rate' clock ticks */
				adj.tick_nsec_inc=1000L*(usec/rate); /* may be up to 'rate' usec of error */

				/*zzx need to check two things: first, that the error time
                   is not greater than our allowable threshold, and
                   second, that the per tick amount is not greater
                   than the amount of time in a tick */
	
                adj.tick_count = rate;
#ifdef VERBOSE_SUPPORTED
				if (verbose) {
					printf("rtc read: Adjusting clock %d000 nsec\n",usec);
					printf("rtc read: (%d nsec per tick over %d ticks)\n",adj.tick_nsec_inc,adj.tick_count);
				}
#endif
				rc= ClockAdjust(CLOCK_REALTIME,&adj,NULL);
#ifdef VERBOSE_SUPPORTED
				if (verbose>1) {
					do {				
						sleep(1);
						ClockAdjust(CLOCK_REALTIME,NULL,&adj);
						printf("rtc adj: tick_count=%d\n",adj.tick_count);
					} while (adj.tick_count!=0);
                }
#endif
				}
#endif   /* __QNXNTO__ defined */
#endif   /* SLOWADJUST is supported */

			if(rc == -1)
#ifdef __QNXNTO__
				if (ClockTime(CLOCK_REALTIME,&new,NULL)) {
#else
				if(setclock(TIMEOFDAY, &new)) {
#endif
					perror("Can't set system time");
					exit(EXIT_FAILURE);
					}
		}
	}
	exit(0);
}

